import requests
from baangt.base import GlobalConstants as GC
from baangt.TestSteps import Exceptions
import logging
import json

logger = logging.getLogger("pyC")


class ApiHandling:
    def __init__(self):
        """
        This class handles Session Management for API-Calls
        """
        self.session = {}
        self.baseURL = None
        self.endPoint = None
        self.answerStatusCode = None
        self.answerHeaders = None
        self.answerJSON = None

    def getSession(self, sessionNumber=1):
        if not self.session.get(sessionNumber):
            self.session[sessionNumber] = requests.Session()
        return self.session[sessionNumber]

    def getNewSession(self, sessionNumber=None):
        if not sessionNumber:
            # fixme: Get highest Session Number, add 1, etc.
            logger.critical("Method not implemented - aborting")
            raise Exceptions.baangtTestStepException("Method not properly implemented")
        else:
            self.session[sessionNumber] = requests.session()

    def getURL(self, url=None, sessionNumber=1):
        url = self.__retrieveURL(url)
        answer = self.getSession(sessionNumber=sessionNumber).get(url=url)

        self.__saveAnswerLocally(answer=answer, sessionNumber=sessionNumber)

        return self.session[sessionNumber].answerStatusCode, \
               self.session[sessionNumber].answerJSON, \
               self.session[sessionNumber].answerHeaders

    @staticmethod
    def returnTestCaseStatus(status_code):
        if status_code < 300:
            return GC.TESTCASESTATUS_SUCCESS
        else:
            return GC.TESTCASESTATUS_ERROR

    def __saveAnswerLocally(self, answer, sessionNumber):
        self.session[sessionNumber].answerStatusCode = answer.status_code
        self.session[sessionNumber].answerJSON = answer.json()
        self.session[sessionNumber].answerHeaders = dict(answer.headers)

    def postURL(self, url=None, content=None, sessionNumber=1):
        url = self.__retrieveURL(url)
        logger.info(f"POST to url: {url} with content: {content}")
        content = ApiHandling.__cnvtString2Dict(content)
        if isinstance(content, str):
            # Convert content into JSON-Format, if leading and trailing {}
            if content[0] == "{" and content[-1] == "}":
                content = json.loads(content)
        answer = self.getSession(sessionNumber=sessionNumber).post(url=url, data=content)

        self.__saveAnswerLocally(answer=answer, sessionNumber=sessionNumber)

        return self.session[sessionNumber].answerStatusCode, \
               self.session[sessionNumber].answerJSON, \
               self.session[sessionNumber].answerHeaders

    @staticmethod
    def __cnvtString2Dict(content):
        if isinstance(content, str):
            # Convert content into JSON-Format, if leading and trailing {}
            if content[0] == "{" and content[-1] == "}":
                content = json.loads(content)

        return content

    def setLoginData(self, userName, password, sessionNumber=1):
        self.session[sessionNumber].auth = (userName, password)

    def setHeaders(self, sessionNumber=1, setHeaderData=None):
        setHeaderData = ApiHandling.__cnvtString2Dict(setHeaderData)
        self.session[sessionNumber].headers.update(setHeaderData)

    def tearDown(self, sessionNumber=None):
        if not sessionNumber:
            for key, value in self.session.items():
                self.session[key].close()

    def __retrieveURL(self, url):
        """

        @param url: Importing parameter - maybe an Endpoint-URL
        @return: either the passed URL or the combination of BaseURL + Endpoint.
        """
        if url:
            return url

        url = self.baseURL + self.endPoint
        return url

    def setBaseURL(self, url):
        self.baseURL = url

    def setEndPoint(self, endpoint):
        self.endPoint = endpoint



import logging
from datetime import datetime
import pathlib
import sys
import os
from pathlib import Path
from pluggy import HookspecMarker, HookimplMarker, PluginManager
from baangt.base.PathManagement import ManagedPaths

hook_spec = HookspecMarker("baangt")
hook_impl = HookimplMarker("baangt")
plugin_manager = PluginManager("baangt")

from baangt.hookSpecs import baangtHookSpec

plugin_manager.add_hookspecs(baangtHookSpec)

from baangt.base.TestRun.hookImpls import TestRunHookImpl
from baangt.base.BrowserHandling.hookImpls import BrowserDriverHookImpl
from baangt.base.Timing.hookImpls import TimingHookImpl
from baangt.base.ExportResults.hookImpls import \
    (ExportResultsHookImpl, ExcelSheetHelperFunctionsHookImpl, ExportTimingHookImpl)

plugin_manager.register(plugin=TestRunHookImpl())
plugin_manager.register(plugin=BrowserDriverHookImpl())
plugin_manager.register(plugin=TimingHookImpl())
plugin_manager.register(plugin=ExportResultsHookImpl())
plugin_manager.register(plugin=ExcelSheetHelperFunctionsHookImpl())
plugin_manager.register(plugin=ExportTimingHookImpl())

# fixme: Parameter für Logfile should include stage and browser()
managedPaths = ManagedPaths()
logFilename:pathlib.Path = Path(managedPaths.getLogfilePath())
pathlib.Path(logFilename).mkdir(parents=True, exist_ok=True)
logFilename = logFilename.joinpath(datetime.now().strftime("%Y%m%d_%H%M%S") + '.log')

print(f"Logfile used: {logFilename}")

# Bit more advanced logging
logger = logging.getLogger('pyC')
logger.setLevel(logging.DEBUG)
# create file handler which logs even debug messages
fileHandler = logging.FileHandler(logFilename, encoding ="UTF-8")
fileHandler.setLevel(level=logging.DEBUG)
# create console handler with a higher log level
channelHandler = logging.StreamHandler()
channelHandler.setLevel(logging.INFO)
# create formatter and add it to the handlers
formatter = logging.Formatter('%(asctime)s _ %(levelname)s _ %(module)s _ %(funcName)s : %(message)s')
channelHandler.setFormatter(formatter)
fileHandler.setFormatter(formatter)
# add the handlers to logger
logger.addHandler(channelHandler)
logger.addHandler(fileHandler)

# Some general debug info:
logger.debug(f"Value of sys._MEIPASS {getattr(sys, '_MEIPASS', 'None')}")
logger.debug(f"Value of __file__: {__file__}")
logger.debug(f"Value of sys.executable: {sys.executable}")
logger.debug(f"Value of sys.argv[0]: {sys.argv[0]}")
logger.debug(f"Value of sys.path: {sys.path}")
logger.debug(f"Value of os.path: {os.path}")
logger.debug(f"Value of os.getcwd(): {os.getcwd()}")
logger.debug(f"Value of Path.cwd(): {Path.cwd()}")

if hasattr(sys, 'frozen') and hasattr(sys, '_MEIPASS'):
    logger.debug('running in a PyInstaller bundle')
else:
    logger.debug('running in a normal Python process')




from baangt.base.TestRun.ClassesForObjects import ClassesForObjects
import baangt.base.GlobalConstants as GC
import logging

logger = logging.getLogger("pyC")

class TestRunUtils():
    def __init__(self):
        self.testRunAttributes = {}

    def setCompleteTestRunAttributes(self, testRunName:str, testRunAttributes: dict):
        self.testRunAttributes[testRunName] = testRunAttributes

    def getCompleteTestRunAttributes(self, testRunName):
        logger.info('get into getCompleteTestRunAttributes, testRunName is {}'.format(testRunName))
        return self.testRunAttributes[testRunName][GC.KWARGS_TESTRUNATTRIBUTES]

    def getSequenceByNumber(self, sequence, testRunName):
        return self.testRunAttributes[testRunName][GC.KWARGS_TESTRUNATTRIBUTES][GC.STRUCTURE_TESTCASESEQUENCE].get(sequence)

    def getTestCaseByNumber(self, sequence, testcaseNumber):
        return sequence[1][GC.STRUCTURE_TESTCASE].get(testcaseNumber) or \
               sequence[1][GC.STRUCTURE_TESTCASE].get(int(testcaseNumber)) or \
               sequence[1][GC.STRUCTURE_TESTCASE].get(str(testcaseNumber))

    def getTestStepByNumber(self, testCase, testStepNumber):
        return testCase[2][GC.STRUCTURE_TESTSTEP].get(testStepNumber) or \
               testCase[2][GC.STRUCTURE_TESTSTEP].get(int(testStepNumber)) or \
               testCase[2][GC.STRUCTURE_TESTSTEP].get(str(testStepNumber))


    def replaceClasses(self, testRunName, classes:ClassesForObjects, ):
        """
        Apart from what is defined in the TestRunDefintion (XLSX or "guessed" by TestRunExcelImporter) the user
        may give new class names in the globals file.

        We shall replace all corresponding appearances with those new class names

        :param classes:
        :return: No return parameter. Changes are kept internally
        """
        lTestRun = self.testRunAttributes[testRunName]

        try:
            for n in range(1,999):
                lSequence = self.getSequenceByNumber(n, testRunName=testRunName)
                if lSequence:
                    lSequence[0] = classes.testCaseSequenceMaster
                    lSequence[1]["SequenceClass"] = classes.testCaseSequenceMaster
                else:
                    break
        except BaseException as e:
            pass # This was the last TestCaseSequence within this test case.

    def replaceGlobals(self, globals):
        """
        Will go through all testcase-Settings and replace values with values from global settings, if matched
        """
        for key, value in globals.items():
            if not "TC." in key[0:3]:
                continue
            if not isinstance(value, bool):
                if len(value) == 0:
                    continue
            self.testRunAttributes = TestRunUtils._recursive_replace(self.testRunAttributes, key.replace("TC.",""), value)

    @staticmethod
    def _recursive_replace(dictToBeReplaced, lKey, lValue):
        if isinstance(dictToBeReplaced, list):
            for entry in dictToBeReplaced:
                TestRunUtils._recursive_replace(entry, lKey, lValue)
        elif isinstance(dictToBeReplaced, dict):
            if lKey in dictToBeReplaced:
                logger.info(f"Due to Globals replaced value {dictToBeReplaced[lKey]} of {lKey} with value {lValue}")
                dictToBeReplaced[lKey] = lValue
            for k,v in dictToBeReplaced.items():
                TestRunUtils._recursive_replace(v, lKey, lValue)
        else:
            pass
        return dictToBeReplaced



from sqlalchemy import Column, String, Integer, DateTime, Boolean, Table, ForeignKey, LargeBinary
#from sqlalchemy.types import Binary, TypeDecorator
from sqlalchemy.orm import relationship
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
import os
import uuid
from baangt.base.PathManagement import ManagedPaths


managedPaths = ManagedPaths()
DATABASE_URL = os.getenv('DATABASE_URL') or 'sqlite:///'+str(managedPaths.derivePathForOSAndInstallationOption().joinpath('testrun.db'))

engine = create_engine(DATABASE_URL)

base = declarative_base()

#
# UUID as bytes
#
def uuidAsBytes():
	return uuid.uuid4().bytes

#
# Testrun models
#

class TestrunLog(base):
	#
	# summary on Testrun results
	#
	__tablename__ = "testruns"
	# columns
	id = Column(LargeBinary, primary_key=True, default=uuidAsBytes)
	testrunName = Column(String, nullable=False)
	logfileName = Column(String, nullable=False)
	startTime = Column(DateTime, nullable=False)
	endTime = Column(DateTime, nullable=False)
	dataFile = Column(String, nullable=True)
	statusOk = Column(Integer, nullable=False)
	statusFailed = Column(Integer, nullable=False)
	statusPaused = Column(Integer, nullable=False)
	# relationships
	globalVars = relationship('GlobalAttribute')
	testcase_sequences = relationship('TestCaseSequenceLog')

	@property
	def recordCount(self):
		return self.statusOk + self.statusPaused + self.statusFailed

	@property
	def duration(self):
		return (self.endTime - self.startTime).seconds
	

	def __str__(self):
		return str(uuid.UUID(bytes=self.id))

	def to_json(self):
		return {
			'id': str(self),
			'Name': self.testrunName,
			'Summary': {
				'TestRecords': self.recordCount,
				'Successful': self.statusOk,
				'Paused': self.statusPaused,
				'Error': self.statusFailed,
				'LogFile': self.logfileName,
				'StartTime': self.startTime.strftime('%H:%M:%S'),
				'EndTime': self.endTime.strftime('%H:%M:%S'),
				'Duration': str(self.endTime - self.startTime),
			},
			'GlobalSettings': {gv.name: gv.value for gv in self.globalVars},
			'TestSequences': [tsq.to_json() for tsq in self.testcase_sequences],
		}



class GlobalAttribute(base):
	#
	# global vars
	#
	__tablename__ = 'globals'
	# columns
	id = Column(Integer, primary_key=True)
	name = Column(String, nullable=False)
	value = Column(String, nullable=True)
	testrun_id = Column(LargeBinary, ForeignKey('testruns.id'), nullable=False)
	# relationships
	testrun = relationship('TestrunLog', foreign_keys=[testrun_id])


#
# Test Case Sequence models
#

class TestCaseSequenceLog(base):
	#
	# TestCase Sequence
	#

	__tablename__ = 'testCaseSequences'
	# columns
	id = Column(LargeBinary, primary_key=True, default=uuidAsBytes)
	testrun_id = Column(LargeBinary, ForeignKey('testruns.id'), nullable=False)
	# relationships
	testrun = relationship('TestrunLog', foreign_keys=[testrun_id])
	testcases = relationship('TestCaseLog')

	def __str__(self):
		return str(uuid.UUID(bytes=self.id))

	def to_json(self):
		return {
			'id': str(self),
			'TestCases': [tc.to_json() for tc in self.testcases],
		}


#
# Test Case models
#

class TestCaseLog(base):
	#
	# TestCase results
	#
	__tablename__ = 'testCases'
	# columns
	id = Column(LargeBinary, primary_key=True, default=uuidAsBytes)
	testcase_sequence_id = Column(LargeBinary, ForeignKey('testCaseSequences.id'), nullable=False)
	# relationships
	testcase_sequence = relationship('TestCaseSequenceLog', foreign_keys=[testcase_sequence_id])
	fields = relationship('TestCaseField')
	networkInfo = relationship('TestCaseNetworkInfo')

	def __str__(self):
		return str(uuid.UUID(bytes=self.id))

	def to_json(self):
		return {
			'id': str(self),
			'Parameters': {pr.name: pr.value for pr in self.fields},
			'NetworkInfo': [nw.to_json() for nw in self.networkInfo],
		}


class TestCaseField(base):
	#
	# field for a TestCase results
	#
	__tablename__ = 'testCaseFields'
	# columns
	id = Column(Integer, primary_key=True)
	name = Column(String, nullable=False)
	value = Column(String, nullable=True)
	testcase_id = Column(LargeBinary, ForeignKey('testCases.id'), nullable=False)
	# relationships
	testcase = relationship('TestCaseLog', foreign_keys=[testcase_id])

class TestCaseNetworkInfo(base):
	#
	# network info for a TestCase
	#
	__tablename__ = 'networkInfo'
	# columns
	id = Column(Integer, primary_key=True)
	browserName = Column(String, nullable=True)
	status = Column(Integer, nullable=True)
	method = Column(String, nullable=True)
	url = Column(String, nullable=True)
	contentType = Column(String, nullable=True)
	contentSize = Column(Integer, nullable=True)
	headers = Column(String, nullable=True)
	params = Column(String, nullable=True)
	response = Column(String, nullable=True)
	startDateTime = Column(DateTime, nullable=True)
	duration = Column(Integer, nullable=True)
	testcase_id = Column(LargeBinary, ForeignKey('testCases.id'), nullable=True)
	# relationships
	testcase = relationship('TestCaseLog', foreign_keys=[testcase_id])

	def to_json(self):
		return {
			'BrowserName': self.browserName,
			'Status': self.status,
			'Method': self.method,
			'URL': self.url,
			'ContentType': self.contentType,
			'ContentSize': self.contentSize,
			'Headers': self.headers,
			'Params': self.params,
			'Response': '',#self.response,
			'StartDateTime': str(self.startDateTime),
			'Duration': self.duration,
		}



# create tables
#if __name__ == '__main__':
base.metadata.create_all(engine)


import os
from selenium.webdriver.common.by import By
from selenium.common.exceptions import *
from selenium.webdriver.common import keys
from baangt.base import GlobalConstants as GC
from baangt.base.Timing.Timing import Timing
from baangt.TestSteps import Exceptions
from baangt.base.DownloadFolderMonitoring import DownloadFolderMonitoring
from baangt.base.BrowserHandling.WebdriverFunctions import WebdriverFunctions as webDrv
from baangt.base.BrowserHandling.BrowserHelperFunction import BrowserDriverData
from baangt.base.BrowserHandling.BrowserHelperFunction import BrowserHelperFunction as helper
from baangt.base.Utils import utils
from baangt.base.ProxyRotate import ProxyRotate
from http import HTTPStatus
import uuid
import time
import logging
from pathlib import Path
#import json
import sys
import platform
#import ctypes
import requests
from baangt.base.PathManagement import ManagedPaths
from baangt.base.RuntimeStatistics import Statistic


logger = logging.getLogger("pyC")

class BrowserDriver:
    """
    The main class for baangt-Elements to interact with a browser.
    Main Methods:
    - createNewBrowser: Create one instance of a Browser
    - findBy*-Methods: e.g. findByAndClick
    - URL-Methods: To navigate to an URL
    - handleIframe and handleWindow: To navigate between Windows (=tabs) and Iframes
    - javaScript: to pass JS directly to the browser
    - takeScreenshot: yes, that.
    """

    def __init__(self, timing=None, screenshotPath=None, statistics=None):
        self.iFrame = None
        self.element = None                 
        self.browserData = BrowserDriverData(locatorType=None, locator=None,
                                             driver=webDrv.BROWSER_DRIVERS[GC.BROWSER_FIREFOX])
        self.slowExecution = False
        self.slowExecutionTimeoutInSeconds = 1
        self.downloadFolder = None
        self.downloadFolderMonitoring = None
        self.randomProxy = None
        self.zoomFactorDesired = None                     # Desired zoom factor for this page
        self.browserName = None
        # Reference to Selenium "HTML" in order to track page changes. It is set on every interaction with the page
        self.html = None
        self.managedPaths = ManagedPaths()
        self.statistics = Statistic()

        if timing:
            self.timing = timing
            self.takeTime = timing.takeTime
        else:
            self.timing = Timing()
            self.takeTime = self.timing.takeTime

        if not screenshotPath or screenshotPath == "":
            self.screenshotPath = self.managedPaths.getOrSetScreenshotsPath()
        else:
            self.screenshotPath = screenshotPath

    def sleep(self, seconds):
        time.sleep(seconds)

    def createNewBrowser(self, mobileType=None, mobileApp = None, desired_app = None, mobile_app_setting = None,
                         browserName=GC.BROWSER_FIREFOX,
                         desiredCapabilities={}, randomProxy=None, **kwargs):
        """
        Will find the specified executables of the desired browser and start it with the given capabilities.

        @param browserName: one of GC_BROWSER_*-Browsernames, e.g. GC_BROWSER_FIREFOX
        @param desiredCapabilities: DICT of desiredCapabilities for this browser
        @param kwargs: Currently (Jan2020) not used
        """
        self.takeTime("Browser Start")
        self.randomProxy = randomProxy
        self.browserName = browserName

        lCurPath = Path(self.managedPaths.getOrSetDriverPath())

        if browserName in webDrv.BROWSER_DRIVERS:

            browserProxy = kwargs.get('browserProxy')
            browserInstance = kwargs.get('browserInstance', 'unknown')

            if utils.anything2Boolean(mobileType):
                self.browserData.driver = self._mobileConnectAppium(browserName, desired_app, mobileApp, mobile_app_setting)
            elif GC.BROWSER_FIREFOX == browserName:
                self.browserData.driver  = self._browserFirefoxRun(browserName, lCurPath, browserProxy, randomProxy, desiredCapabilities)
                helper.browserHelper_startBrowsermobProxy(browserName=browserName, browserInstance=browserInstance, browserProxy=browserProxy)
            elif GC.BROWSER_CHROME == browserName:
                self.browserData.driver = self._browserChromeRun(browserName, lCurPath, browserProxy, randomProxy, desiredCapabilities)
                helper.browserHelper_startBrowsermobProxy(browserName=browserName, browserInstance=browserInstance, browserProxy=browserProxy)
            elif GC.BROWSER_EDGE == browserName:
                self.browserData.driver = webDrv.BROWSER_DRIVERS[browserName](executable_path = helper.browserHelper_findBrowserDriverPaths(GC.EDGE_DRIVER))
            elif GC.BROWSER_SAFARI == browserName:
                # SAFARI doesn't provide any options, but desired_capabilities.
                # Executable_path = the standard safaridriver path.
                if len(desiredCapabilities) == 0:
                    desiredCapabilities = {}
                self.browserData.driver = webDrv.BROWSER_DRIVERS[browserName](desired_capabilities=desiredCapabilities)
            elif GC.BROWSER_REMOTE == browserName:
                self.browserData.driver = webDrv.BROWSER_DRIVERS[browserName](options = webDrv.webdriver_createBrowserOptions(browserName=browserName,
                                                        desiredCapabilities=desiredCapabilities),
                                                        command_executor=GC.REMOTE_EXECUTE_URL,
                                                        desired_capabilities=desiredCapabilities)
            else:
                # TODO add exception, this code should never be reached
                pass
        elif GC.BROWSER_REMOTE_V4 == browserName:
            desired_capabilities, seleniumGridIp, seleniumGridPort = helper.browserHelper_setSettingsRemoteV4(desiredCapabilities)

            if desired_capabilities['browserName'] == 'firefox':
                browserExecutable = helper.browserHelper_getBrowserExecutable(GC.BROWSER_FIREFOX)
                self._downloadDriverCheck(browserExecutable, lCurPath, GC.BROWSER_FIREFOX)
            elif desired_capabilities['browserName'] == 'chrome':
                browserExecutable = helper.browserHelper_getBrowserExecutable(GC.BROWSER_CHROME)
                self._downloadDriverCheck(browserExecutable, lCurPath, GC.BROWSER_CHROME)

            serverUrl = 'http://' + seleniumGridIp + ':' + seleniumGridPort
            self.browserData.driver = webDrv.BROWSER_DRIVERS[GC.BROWSER_REMOTE](command_executor=serverUrl, desired_capabilities=desiredCapabilities)
        else:
            raise SystemExit("Browsername unknown")

        if self.downloadFolder:
            self.downloadFolderMonitoring = DownloadFolderMonitoring(self.downloadFolder)

        self.takeTime("Browser Start")

    def _downloadDriverCheck(self, executable, lCurPath, browserName):
        lCurPath = lCurPath.joinpath(executable)

        if not (os.path.isfile(str(lCurPath))):
            self.downloadDriver(browserName)

    def _browserChromeRun(self, browserName, lCurPath, browserProxy, randomProxy, desiredCapabilities):
        executable = helper.browserHelper_getBrowserExecutable(browserName)
        self._downloadDriverCheck(executable, lCurPath, browserName)

        lOptions = webDrv.webdriver_createBrowserOptions(browserName=browserName,
                                                        desiredCapabilities=desiredCapabilities,
                                                        browserMobProxy=browserProxy,
                                                        randomProxy=randomProxy)

        self.downloadFolder=webDrv.getDownloadFolderFromChromeOptions(options=lOptions)

        return webDrv.BROWSER_DRIVERS[browserName](
            chrome_options = lOptions,
            executable_path = helper.browserHelper_findBrowserDriverPaths(executable),
            service_log_path = os.path.join(self.managedPaths.getLogfilePath(), 'chromedriver.log')
        )

    def _browserFirefoxRun(self, browserName, lCurPath, browserProxy, randomProxy, desiredCapabilities):
        executable = helper.browserHelper_getBrowserExecutable(browserName)
        self._downloadDriverCheck(executable, lCurPath, browserName)

        profile = webDrv.webdriver_setFirefoxProfile(browserProxy, randomProxy)
        self.downloadFolder = webDrv.getDownloadFolderFromProfile(profile)
        logger.debug(f"Firefox Profile as follows:{profile.userPrefs}")

        return webDrv.BROWSER_DRIVERS[browserName](
            options = webDrv.webdriver_createBrowserOptions(browserName=browserName, desiredCapabilities=desiredCapabilities),
            executable_path = helper.browserHelper_findBrowserDriverPaths(executable),
            firefox_profile = profile,
            service_log_path = os.path.join(self.managedPaths.getLogfilePath(), 'geckodriver.log')
            )


    @staticmethod
    def _mobileConnectAppium(browserName, desired_app, mobileApp, mobile_app_setting):
        validSettings = False
        desired_cap = desired_app

        if desired_app[GC.MOBILE_PLATFORM_NAME] == "Android":
            validSettings = True
            if utils.anything2Boolean(mobileApp):
                desired_cap['app'] = mobile_app_setting[GC.MOBILE_APP_URL]
                desired_cap['appPackage'] = mobile_app_setting[GC.MOBILE_APP_PACKAGE]
                desired_cap['appActivity'] = mobile_app_setting[GC.MOBILE_APP_ACTIVITY]
            else:
                desired_cap['browserName'] = browserName
                desired_cap['chromedriverExecutable'] = mobile_app_setting[GC.MOBILE_APP_BROWSER_PATH]
                desired_cap['noReset'] = False
            
        elif desired_app[GC.MOBILE_PLATFORM_NAME] == "iOS":
            validSettings = True
            if utils.anything2Boolean(mobileApp):
                desired_cap['automationName'] = 'XCUITest'
                desired_cap['app'] = mobile_app_setting[GC.MOBILE_APP_URL]
            else:
                desired_cap['browserName'] = 'safari'
            
        if validSettings:
            return webDrv.BROWSER_DRIVERS[GC.BROWSER_APPIUM](GC.REMOTE_EXECUTE_URL, desired_cap)
        else:
            return None

    def closeBrowser(self):
        self.statistics.update_teststep()
        try:
            if self.browserData.driver:
                self.browserData.driver.quit()
                self.browserData.driver = None
        except Exception as ex:
            pass  # If the driver is already dead, it's fine.


    def refresh(self):
        self.browserData.driver.execute_script("window.location.reload()")

    def takeScreenshot(self, screenShotPath=None):
        driver = self.browserData.driver
        # Filename must have ".png" inside
        lFile = str(uuid.uuid4()) + ".png"

        if screenShotPath:
            lFile = Path(screenShotPath).joinpath(lFile)
        else:
            lFile = Path(self.screenshotPath).joinpath(lFile)

        try:
            lFile = str(lFile)
            driver.save_screenshot(lFile)
            helper.browserHelper_log(logging.DEBUG, f"Stored Screenshot: {lFile}", self.browserData)
        except Exception as e:
            helper.browserHelper_log(logging.INFO, f"Screenshot not possible. Error: {e}", self.browserData)
            lFile = None

        return lFile

    def handleIframe(self, iframe=None):
        """
        Give an IFRAME and it will try to go into.
        If you're inside an iframe it will go out of the iframe
        """
        self.statistics.update_teststep()
        if iframe:
            self.browserData.locatorType="XPATH"
            self.browserData.locator=iframe
            helper.browserHelper_log(logging.DEBUG, "Going into Iframe: ", self.browserData, **{"iframe": iframe})
            # frame_to_be_availble_and_switch_to_it doesn't work.
            mustEnd = time.time() + 30
            while time.time() < mustEnd:
                try:
                    self.browserData.driver.switch_to.default_content()
                    self.iFrame = self.browserData.driver.switch_to.frame(iframe)
                    break
                except WebDriverException as e:
                    helper.browserHelper_log(logging.DEBUG, f"IFrame {iframe} not there yet - waiting 1 second", self.browserData)
                    time.sleep(1)

            if time.time() > mustEnd:
                raise TimeoutError

        elif self.iFrame:
            helper.browserHelper_log(logging.DEBUG, f"Leaving Iframe: {self.iFrame}", self.browserData)
            self.browserData.driver.switch_to.default_content()
            self.iFrame = None
        else:
            # TODO add exception, this code should never be reached
            pass

    def handleWindow(self, windowNumber=None, function=None, timeout=20):
        """
        Interations with Windows (=BrowserTabs).

        @param windowNumber: Number of the windowHandle inside this browser session (0 = startwindow(=Tab), 1=Next window
        @param function: "CLOSE", "CLOSEALL"
        """
        self.statistics.update_teststep()
        if function:
            if "close" == function.lower():
                self.browserData.driver.close()
                self.browserData.driver.switch_to.window(self.browserData.driver.window_handles[0])
            elif "closeall" in function.lower():
                exceptHandles = function.lower().replace("closeall", "")
                exceptHandles = exceptHandles.replace("-", "")
                # WindowHandles based on 0.. Value "let 2 windows open" means to close everything except 0 and 1:
                exceptHandles = int(exceptHandles.strip()) - 1
                try:
                    len(self.browserData.driver.window_handles)
                except BaseException as e:
                    logger.error(f"Tried to get amount of windows. Threw error {e}. Most probably browser crashed")
                    raise Exceptions.baangtTestStepException(f"Tried to get amount of windows. "
                                                             f"Threw error {e}. Most probably browser crashed")
                for windowHandle in self.browserData.driver.window_handles[-1:exceptHandles:-1]:
                    try:
                        self.browserData.driver.switch_to.window(windowHandle)
                        self.browserData.driver.close()
                    except NoSuchWindowException as e:
                        # If the window is already closed, it's fine. Don't do anything
                        pass
                try:
                    self.browserData.driver.switch_to.window(self.browserData.driver.window_handles[exceptHandles])
                except IndexError as e:
                    raise Exceptions.baangtTestStepException(f"Seems like the browser crashed. Main-Window lost")
            else:
                # TODO Wrong function, add exception
                pass
        else:
            success = False
            duration = 0
            while not success and duration < timeout:
                try:
                    self.browserData.driver.switch_to.window(self.browserData.driver.window_handles[windowNumber])
                    success = True
                    continue
                except Exception as e:
                    logger.debug(f"Tried to switch to Window {windowNumber} but it's not there yet")

                time.sleep(1)
                duration += 1

            if not success:
                raise Exceptions.baangtTestStepException(f"Window {windowNumber} doesn't exist after timeout {timeout}")

    def findByAndWaitForValue(self, id=None, css=None, xpath=None, class_name=None, iframe=None, timeout=20, optional=False):
        """

        @param id: ID of the element
        @param css: CSS-Locator
        @param xpath: XPATH-Locator
        @param class_name: Class-Name
        @param iframe: Iframe to use (use only if changed. If you set an iframe before, you don't need to set it again!)
        @param timeout: Timeout in Seconds before raising an error or returning back (depending on "optional")
        @param optional: If set to "True" and the operation can not be executed, just a log entry is written but no error raised
        @return: the text of the element, if element was found
        """
        self.statistics.update_teststep()
        self.element = None
        returnValue = None
        start = time.time()
        duration = 0

        while not self.element and duration < timeout:
            self.element, self.html = self.findBy(id=id, css=css, xpath=xpath, class_name=class_name, iframe=iframe, timeout=timeout / 3,
                            optional=optional)
            time.sleep(0.5)
            duration = time.time() - start

        if self.element:
            try:
                if len(self.element.text) > 0:
                    returnValue = self.element.text
                elif self.element.tag_name == 'input':
                    #  element is of type <input />
                    returnValue = self.element.get_property('value')
                else:
                    returnValue = None
            except Exception as e:
                logger.debug(f"Exception during findByAndWaitForValue, but continuing {str(e)}, "
                             f"Locator: {self.browserData.locatorType} = {self.browserData.locator}")
        else:
            logger.info(f"Couldn't find value for element {self.browserData.locatorType}:{self.browserData.locator}")

        return returnValue

    def findByAndSetText(self, id=None, css=None, xpath=None, class_name=None, value=None, iframe=None,
                         timeout=60, optional=False):
        """
        Please see documentation in findBy and __doSomething
        """
        self.element, self.html = self.findBy(id=id, css=css, xpath=xpath, class_name=class_name, iframe=iframe, timeout=timeout)

        return webDrv.webdriver_doSomething(GC.CMD_SETTEXT, self.element, value=value, timeout=timeout, optional=optional, browserData = self.browserData)

    def slowExecutionToggle(self, newSlowExecutionWaitTimeInSeconds=None):
        """
        SlowExecution can be set in globals or by the teststep. It's intended use is debugging or showcasing a testcases
        functionality.

        @param newSlowExecutionWaitTimeInSeconds: Optional. If set, it will change the default value of WaitTime, when SlowExecution is active
        @return: Returns the state of sloeExecution toggle after toggling was done.
        """

        if newSlowExecutionWaitTimeInSeconds:
            self.slowExecutionTimeoutInSeconds = newSlowExecutionWaitTimeInSeconds

        return not self.slowExecution

    def findByAndSetTextIf(self, id=None, css=None, xpath=None, class_name=None, value=None, iframe=None,
                           timeout=60, optional=False):
        """
        Helper function to not have to write:
        If <condition>:
            findByAndSetText(locator)

        instead use:
        findByAndSetTextIf(locator, value).

        If value is evaluated into "True" the Text is set.

        """

        if self._isValidKeyValue(value):
            return self.findByAndSetText(id=id, css=css, xpath=xpath, class_name=class_name, value=value, iframe=iframe, timeout=timeout, optional=optional)
        else:
            return False

    def findByAndSetTextValidated(self, id=None, css=None, xpath=None, class_name=None, value=None, iframe=None, timeout=60, retries=5):
        """
        This is a method not recommended to be used regularly. Sometimes (especially with Angular Frontends) it gets
        pretty hard to set a value into a field. Chrome, but also FF will show the value, but the DOM will not have it.
        Ths Method should be your last ressort. Here we try <retries> time to set a value. Then we read the element again
        and compare value to what we'd expect. If value is different and we're less than <retries>-Times, we'll try again.
        """

        tries = 0

        self.element, self.html = self.findBy(id=id, css=css, xpath=xpath, class_name=class_name, iframe=iframe, timeout=timeout)

        while self.element.text != value and self.element.get_property("value") != value and tries < retries:
            helper.browserHelper_log(logging.DEBUG, f"Verified trying of SetText - iteration {tries} of {retries}", self.browserData)

            self.findByAndForceText(id=id, css=css, xpath=xpath, class_name=class_name, iframe=iframe,
                                    value=value, timeout=timeout)

            self.element, self.html = self.findBy(id=id, css=css, xpath=xpath, class_name=class_name, iframe=iframe, timeout=timeout)

            tries += 1

    def submit(self):
        """
        Used for forms to call the standard submit-function (similar to pressing "Enter" in Dialogue)
        @return:
        """
        self.statistics.update_teststep()
        self.element.submit()

    def findByAndClick(self, id=None, css=None, xpath=None, class_name=None, iframe=None, timeout=20, optional=False):
        """
        Execute a Click on an element identified by it's locator.
        @return wasSuccessful says, whether the element was found.
        """

        self.element, self.html = self.findBy(id=id, css=css, xpath=xpath, class_name=class_name, iframe=iframe, timeout=timeout,
                                    optional=optional)

        if not self.element:
            logger.debug("findBy didn't work in findByAndClick")
            return False
        else:
            return webDrv.webdriver_doSomething(GC.CMD_CLICK, self.element, timeout=timeout, optional=optional, browserData = self.browserData)

    def confirmAlertIfAny(self):
        self.statistics.update_teststep()
        try:
            self.browserData.driver.switch_to().alert().accept()
        except Exception as e:
            pass

    @staticmethod
    def _isValidKeyValue(value):
        isValid = False
        if not value:
            pass
        elif len(value) == 0 or str(value) == "0":
            pass
        else:
            isValid = True
        return isValid 

    def findByAndClickIf(self, id=None, css=None, xpath=None, class_name=None, iframe=None, timeout=60,
                         value=None, optional=False):
        """
        Convenience method to not have to write:
        if <condition>:
            findByAndClick(locator)

        instead write:
        findByAndClickIf(locator, value).

        If value is evaluated to "True", the click-event is executed.
        """
        if self._isValidKeyValue(value):
            return self.findByAndClick(id=id, css=css, xpath=xpath, class_name=class_name, iframe=iframe, timeout=timeout, optional=optional)
        else:
            return False

    def findByAndForceText(self, id=None, css=None, xpath=None, class_name=None, value=None,
                           iframe=None, timeout=60, optional=False):
        """
        Convenience Method. Please see documentation in findBy and __doSomething.

        """

        self.element, self.html = self.findBy(id=id, css=css, xpath=xpath, class_name=class_name, iframe=iframe, timeout=timeout)

        if not self.element:
            return False

        return webDrv.webdriver_doSomething(GC.CMD_FORCETEXT, self.element, value=value, timeout=timeout, optional=optional, browserData = self.browserData)

    def setBrowserWindowSize(self, browserWindowSize: str):
        """
        Resized the browser Window to a fixed size
        :param browserWindowSize: String with Widht/Height or Width;Height or Width,height or width x height
               If you want also with leading --
        :return: False, if browser wasn't reset,
                 size-Dict when resize worked.
        """
        self.statistics.update_teststep()
        lIntBrowserWindowSize = browserWindowSize.replace("-","").strip()
        lIntBrowserWindowSize = lIntBrowserWindowSize.replace(";", "/")
        lIntBrowserWindowSize = lIntBrowserWindowSize.replace(",", "/")
        lIntBrowserWindowSize = lIntBrowserWindowSize.replace("x", "/")
        lIntBrowserWindowSize = lIntBrowserWindowSize.replace("*", "/")
        validSize = False

        try:
            width = int(lIntBrowserWindowSize.split("/")[0])
            height = int(lIntBrowserWindowSize.split("/")[1])
        except KeyError as e:
            logger.warning(f"Called with wrong setting: {browserWindowSize}. Won't resize browser "
                           f"Can't determine Width/Height.")
        except ValueError as e:
            logger.warning(f"Called with wrong setting: {browserWindowSize}. Won't resize browser "
                           f"Something seems not numeric before conversion: {lIntBrowserWindowSize}")

        try:
            if width == 0 or height == 0:
                logger.warning(f"Called with wrong setting: {browserWindowSize}. Won't resize browser. Can't be 0")
            else:
                validSize = True
        except:
            pass

        if validSize:
            self.browserData.driver.set_window_size(width, height)
            size = self.browserData.driver.get_window_size()
            logger.debug(f"Resized browser window to width want/is: {width}/{size['width']}, "
                        f"height want/is: {height}/{size['height']}")
        else:
            size = False

        return size

    def findNewFiles(self):
        """
        Returns a list of new files from downloadFolderMonitoring since the last call

        :return: List of Files since last call
        """
        self.statistics.update_teststep()
        l_list = self.downloadFolderMonitoring.getNewFiles()
        return l_list

    @staticmethod
    def _setLocator(id, css, xpath, class_name, browserData):
        browserData.locatorType = None
        browserData.locator = None
        if xpath:
            browserData.locatorType = By.XPATH 
            browserData.locator = xpath
        elif css:
            browserData.locatorType = By.CSS_SELECTOR
            browserData.locator = css
        elif class_name:
            browserData.locatorType = By.CLASS_NAME 
            browserData.locator = class_name
        elif id:
            browserData.locatorType = By.ID 
            browserData.locator = id
        return browserData

    def findBy(self, id=None, css=None, xpath=None, class_name=None, iframe=None, timeout=60, loggingOn=True, optional=False):
        """
        chose any single locator (ID, CSS, XPATH, CLASS_NAME) to identify an element within the page. if slowExectuion
        is set, we'll pause for slowExecutionTimeoutInSeconds.

        @param id: ID of the element
        @param css: CSS-Locator
        @param xpath: XPATH
        @param class_name: Class-Name
        @param iframe: Name of an Iframe. Use only, if you didn't set the Iframe previously already!
        @param timeout: How many seconds shall we try/retry, default = 60 Seconds
        @param loggingOn: Shall this request be logged? Default = Yes
        @param optional: If set to true and within Timeout we can't find the element, we just return this info. If set to False (=default), an Exception is raised
        @return: True if element was located, False if element couldn't be found.
        """
        self.statistics.update_teststep()

        if self.slowExecution:
            time.sleep(self.slowExecutionTimeoutInSeconds)

        if iframe:
            self.handleIframe(iframe)

        # Set class variables for potential logging of problems.
        self.browserData = self._setLocator(id, css, xpath, class_name, self.browserData)

        if loggingOn:
            logger.debug(f"Locating Element {self.browserData.locatorType} = {self.browserData.locator}")

        self.element, self.html = webDrv.webdriver_tryAndRetry(self.browserData, timeout=timeout, optional=optional)

        if not self.element and not optional:
            raise Exceptions.baangtTestStepException(f"Element {self.browserData.locatorType} = {self.browserData.locator} could not be found "
                                                     f"within timeout of {timeout}")
        return self.element, self.html

    def getURL(self):
        """

        @return: the current URL/URI of the current Tab of the current Browser
        """
        self.statistics.update_teststep()
        return self.browserData.driver.current_url


    def findWaitNotVisible(self, css=None, xpath=None, id=None, timeout=90, optional=False):
        """
        You'd use this method when you wait for an element to disappear, for instance Angular Spinner or a popup
        to disapear before you continue with your script in the main screen.

        """
        self.statistics.update_teststep()
        logger.debug(f"Waiting for Element to disappear: XPATH:{xpath}, timeout: {timeout}")
        time.sleep(0.5)

        stillHere = True
        elapsed = 0
        begin = time.time()

        while stillHere and elapsed < timeout:
            try:
                if xpath:
                    self.element = self.browserData.driver.find_element_by_xpath(xpath)
                elif id:
                    self.element = self.browserData.driver.find_element_by_id(id)
                elif css:
                    self.element = self.browserData.driver.find_element_by_css_selector(css)
                time.sleep(0.2)
                elapsed = time.time() - begin
            except Exception as e:
                # Element gone - exit
                stillHere = False
                helper.browserHelper_log(logging.DEBUG, f"Element was gone after {format(elapsed, '.2f')} seconds", self.browserData)

        if not stillHere:
            raise Exceptions.baangtTestStepException(
                f"Element still here after {timeout} seconds. Locator: xpath={xpath}, id={id}")
        
        return stillHere

    def checkLinks(self):
        """
        For the current page we'll check all links and return result in format
        <status_code> <link_that_was_checked>

        :return: List of checked links
        """
        self.statistics.update_teststep()
        lResult = []
        links = self.browserData.driver.find_elements_by_css_selector("a")
        logger.debug(f"Checking links on page {self.browserData.driver.current_url}")
        for link in links:
            lHref = link.get_attribute("href")
            if not lHref:
                continue
            if lHref.startswith("mailto"):
                pass
            else:
                try:
                    r = requests.head(lHref)
                    lResult.append([r.status_code, lHref])
                    logger.debug(f"Result was: {r.status_code}, {lHref}")
                except requests.exceptions.InvalidURL as e:
                    lResult.append([HTTPStatus.INTERNAL_SERVER_ERROR, f"Invalid URL: {lHref}"])  
                except requests.exceptions.ConnectionError as e:
                    lResult.append([HTTPStatus.INTERNAL_SERVER_ERROR, f"HTTP connection error: {lHref}"])
                except requests.exceptions.MissingSchema as e:
                    lResult.append([HTTPStatus.INTERNAL_SERVER_ERROR, f"Missing Schema - invalid URL: {lHref}"])

        return lResult

    def waitForElementChangeAfterButtonClick(self, timeout=5):
        """
        Wait for a stale element (in a good way). Stale means, that the object has changed.

        old element is in self.element
        old locator is in self.browserData.locatorType and self.browserData.locator

        :param timeout:
        :return:
        """

        self.statistics.update_teststep()

        lOldElement = self.element.id
        isValid = False
        lStartOfWaiting = time.time()
        elapsed = 0
        logger.debug("Starting")

        xpath, css, id = utils.setLocatorFromLocatorType(self.browserData.locatorType, self.browserData.locator)

        while not isValid and elapsed < timeout:
            self.element, self.html = self.findBy(xpath=xpath, css=css, id=id, timeout=0.5, optional=True)
            if not self.element:
                # Wonderful. Element is gone
                logger.debug("Old object is not in the page any longer, save to continue")
                isValid = True
            if self.element.id != lOldElement:
                logger.debug("Old element is stale, save to continue")
                isValid = True

            time.sleep(0.2)
            elapsed = time.time() - lStartOfWaiting

        if not isValid:
            # TimeOut Return false
            logger.debug("Old element equal to new element after timeout. Staleness not detected using this method")

        return isValid

    def waitForPageLoadAfterButtonClick(self, timeout=5):
        """
        Problem: If the same XPATH/CSS/ID exists on both pages (the current one, where a button is clicked
                 and the next one, where we now want to interact, then it happens very often, that the element
                 is stale (because it was bound to the current page BEFORE the page-load happened.
        Solution: Wait deliberately until current self.element is stale.
        :param timout: Yeah, you guessed it. The timeout
        :return: True = New page loaded, False = The element didn't get stale within timeout
        """

        # Performance in 5 parallel Runs dropped from 06:50 to 07:51. That's 1 Minute slower
        # 60 Seconds or 10% time lost.
        # For now let it as it is. If users report that as a problem, revisit the subject and
        # e.g. find another way to understand, whether we're still on the same page or not.

        self.statistics.update_teststep()

        if not self.html:
            sys.exit("Something is very wrong! self.html didn't exist when waitForPageLoadAfterButtonClick was called")

        lStartOfWaiting = time.time()
        elapsed = 0
        logger.debug("Starting")

        while elapsed < timeout:
            lHTML = self.browserData.driver.find_element_by_tag_name("html")
            if lHTML != self.html:
                logger.debug("Page was reloaded")
                return True

            time.sleep(0.2)

            elapsed = time.time() - lStartOfWaiting

        logger.debug("No Page reload detected by this method")
        return False    # There was no changed HTML

    def goToUrl(self, url):
        self.statistics.update_teststep()
        helper.browserHelper_log(logging.INFO, f'GoToUrl:{url}', self.browserData)
        try:
            if self.browserName==GC.BROWSER_FIREFOX:
                self.browserData.driver.set_context("content")
            self.browserData.driver.get(url)
            self.setZoomFactor()
        except WebDriverException as e:
            # Use noScreenshot-Parameter as otherwise we'll try on a dead browser to create a screenshot
            helper.browserHelper_log(logging.ERROR, f"Webpage {url} not reached. Error was: {e}", self.browserData, cbTakeScreenshot=self.takeScreenshot)
            helper.browserHelper_setProxyError(self.randomProxy)
            raise Exceptions.baangtTestStepException
        except Exception as e:
            # Use noScreenshot-Parameter as otherwise we'll try on a dead browser to create a screenshot
            helper.browserHelper_log(logging.ERROR, f"Webpage {url} throws error {e}", self.browserData, cbTakeScreenshot=self.takeScreenshot)
            helper.browserHelper_setProxyError(self.randomProxy)
            raise Exceptions.baangtTestStepException(url, e)


    def goBack(self):
        """
        Method to go 1 step back in current tab's browse history
        @return:
        """
        self.statistics.update_teststep()
        try:
            self.javaScript("window.history.go(-1)")
        except Exception as e:
            helper.browserHelper_log(logging.WARNING, f"Tried to go back in history, didn't work with error {e}", self.browserData)

    def javaScript(self, jsText, *args):
        """Execute a given JavaScript in the current Session"""
        self.statistics.update_teststep()
        self.browserData.driver.execute_script(jsText, *args)
        
    def _zoomFirefox(self, lZoomKey, lHitKeyTimes):
        try:
            lWindow = self.browserData.driver.find_element_by_tag_name("html")
            # Reset the browser window to 100%:
            if platform.system().lower() == "darwin":
                lWindow.send_keys(keys.Keys.META + "0")
            else:
                lWindow.send_keys(keys.Keys.CONTROL + "0")

            # Now set to desired zoom factor:
            for counter in range(lHitKeyTimes):
                if platform.system().lower() == "darwin":
                    lWindow.send_keys(keys.Keys.META + lZoomKey)
                else:
                    lWindow.send_keys(keys.Keys.CONTROL + lZoomKey)

            logger.debug(f"Adjusted zoom factor of browserwindow to {self.zoomFactorDesired}")
        except Exception as e:
            logger.debug(f"Tried to adjust zoom factor and failed: {e}")
        finally:
            self.browserData.driver.set_context("content")


    def setZoomFactor(self, lZoomFactor=None):
        """
        Will try to set the browser's zoom factor.

        :param lZoomFactor: set with a value. Otherwise existing value will be used (if previously set)
        :return:
        """

        isZoomed = False
        if self.zoomFactorDesired and lZoomFactor:
            self.zoomFactorDesired = int(lZoomFactor)
            if self.browserName == GC.BROWSER_CHROME:
                logger.critical(f"Zoom in Chrome doesn't work. Continuing without zoom")
                return False
                x = self.getURL()
                if x[0:5] == "http:":       # He loaded already something. Too late for us
                    logger.debug("CHROME: Got called to change Zoom level - but already URL loaded. Too late.")
                else:
                    self.browserData.driver.get("chrome://settings/")
                    self.browserData.driver.execute_script(f"chrome.settingsPrivate.setDefaultZoom({self.zoomFactorDesired/100});")
                    logger.debug(f"CHROME: Set default zoom using JS-Method to {self.zoomFactorDesired/100}")
                    isZoomed = True
            elif self.browserName == GC.BROWSER_FIREFOX:
                self.browserData.driver.set_context("chrome")   

                lZoomKey = "+" if self.zoomFactorDesired > 100 else "-"
                # E.g. current = 100. Desired = 67%: 100-67 = 33. 33/10 = 3.3  int(3.3) = 3 --> he'll hit 3 times CTRL+"-"
                lDifference = abs(100 - self.zoomFactorDesired)
                lHitKeyTimes = int(lDifference/10)
                self._zoomFirefox(lZoomKey, lHitKeyTimes)
                isZoomed = True
            else:
                # statement not matched
                pass
        else:
            # statement not matched
            pass

        return isZoomed

    @staticmethod
    def downloadDriver(browserName):
        managedPaths = ManagedPaths()
        path = Path(managedPaths.getOrSetDriverPath())
        logger.debug(f"Trying to download browserDriver for {browserName} into {path}")
        path.mkdir(parents=True, exist_ok=True)
        if browserName == GC.BROWSER_FIREFOX:
            url, isTarFile = helper.browserHelper_getFirefoxFileUrl()
            if isTarFile:
                helper.browserHelper_extractTarDriverFile(url, path, GC.GECKO_DRIVER)
            else:
                helper.browserHelper_unzipDriverFile(url, path, GC.GECKO_DRIVER)
        elif browserName == GC.BROWSER_CHROME:
            url = helper.browserHelper_getChromeFileUrl()
            helper.browserHelper_unzipDriverFile(url, path, GC.CHROME_DRIVER)
        else:
            logger.critical(f"Please download driver for {browserName} manually into folder /browserDrivers")


class AddressCreate:
    """
    This file Define a Singleton Class AddressCreate which will have static
        method returnAddress to return the address in format.


        CountryCode**
        PostlCode**
        CityName
        StreetName
        HouseNumber
        AdditionalData1
        AdditionalData2

        Input: value1 = {'CountryCode':"IN",'CityName':"Mohali"} [optional]
               value2 = "sales_"

        Output: {'sales_CountryCode':"IN","sales_PostalCode":"19883",
              "sales_CityName":"Mohali","sales_StreeName":"32HB",
               "sales_AdditionalData1":"Near Hospital",
               "sales_AdditionData2":"Ajanta Mandis Park"}


        Sample Usage:
                Default Value
                >> AddressCreate.returnAddress()

                Pass value of ('HouseNumber', 'AdditionalData1',
                 'AdditionalData2', 'StreetName', 'CityName', 'PostalCode',
                 'CountryCode' ) to change.

               For example:
                >> Update City Name
                  AddressCreate(value1={"CityName":"Something"})

                  or
                  AddressCreate({"CityName":"Something"})

                >> Update Country Name
                  AddressCreate({"CountryCode":"US"})



    """

    __instance = None

    class __AddressCreate:
        """ Inner class   """

        def __init__(self, value1={}, value2=""):
            addressData = {"HouseNumber": "6",
                           "AdditionalData1": "Near MahavirChowk",
                           "AdditionalData2": "Opposite St. Marish Church",
                           "StreetName": "GoreGaon",
                           "CityName": "Ambala",
                           "PostalCode": "160055",
                           "CountryCode": "India",
                           }
            prefixData = ""  # empty initially
            
            self.value1 = addressData
            self.value2 = prefixData

            # Now update the value
            self.updateValue(value1, value2)

        def updateValue(self, value1, value2):
            """ This function will process value1,value2 and
                prepare dictionary data
            """
            # value1
            try:
                value1 = dict(value1)
                value2 = str(value2)

            except Exception as e:
                print("Error Occured :  ", e)
            # process only dictionary data, ignore everything else
            if isinstance(value1, type({})):
                for key, value in value1.items():
                    if key in self.value1:
                        self.value1[key] = value

            if value2.strip():
                self.value2 = value2.strip()

        def __str__(self):
            """ AddressCreation print information"""
            return "AddressCreate Instance id :{} ".format(id(self))

        def __repr__(self):
            return "AddressCreate Instance id:{}\nvalue1: {}\nvalue2: {}".format(
                id(self), self.value1, self.value2)

    @staticmethod
    def getInstance():
        if AddressCreate.__instance is None:
            AddressCreate.__instance = AddressCreate.__AddressCreate()
        return AddressCreate.__instance

    def __new__(cls, value1={}, value2=""):
        """ Get the instance and update the value """
        if not AddressCreate.__instance:
            # Create new instance
            AddressCreate.__instance = AddressCreate.__AddressCreate(
                value1, value2)
        else:
            # Update old instance
            AddressCreate.__instance.updateValue(value1, value2)
        return AddressCreate.__instance

    @staticmethod
    def returnAddress():
        """This function will use value1 and value2 to prepare structured
            Address Data
        """
        # get the instance
        a = AddressCreate.getInstance()
        data = {}

        for key, val in a.value1.items():
            # prefix_value + addresskey = addresskey_value
            # if prefix = "Sales" and "CountryCode" = "India"
            # then final data will be data['Sales_CountryCode'] = India
            if a.value2:
                # prefix is given
                data[a.value2 + key] = val
            else:
                data[key] = val
        # finally return the processed dataDict

        return data



CUST_TOASTS = "Toasts"
# CUST_TOASTS_ERROR = "ToastsError" --> Replaced by GC.TESTCASEERRORLOG
VIGOGFNUMMER = "GF#"
SAPPOLNR = "SAP Polizzennr"
PRAEMIE = "Prämie"
POLNRHOST = "PolNR Host"
VERMITTLER = "Vermittler"
MANDANT = "Mandant"
DONAU = "DON"
WSTV = "WSTV"
VN = "VN"

CLASSES_TESTCASE = "baangtVIG.CustTestCaseMaster.CustTestCaseMaster"


from datetime import datetime
from sqlalchemy.orm import sessionmaker
from sqlalchemy import desc
from baangt.base.DataBaseORM import engine, TestrunLog
from jinja2 import Environment, FileSystemLoader
import json
import os
import webbrowser


class Reports:

	path_to_reports = 'reports'


	def __init__(self):
		self.created = datetime.now()
		
	@property
	def data(self):
		#
		# fetches TestrunLogs data
		#

		db = sessionmaker(bind=engine)()

		data = {}

		# get Testrun names
		testrun_names = [item.testrunName for 
			item in db.query(TestrunLog).order_by(TestrunLog.testrunName).group_by(TestrunLog.testrunName).all()]

		for name in testrun_names:
			# build items by testruns
			logs = db.query(TestrunLog).order_by(desc(TestrunLog.startTime)).filter_by(testrunName=name).all()[-10:]
			empty = max(0, 10-len(logs))

			# build charts
			charts = {}

			charts['figures'] = {
				'records': logs[-1].recordCount,
				'successful': logs[-1].statusOk,
				'error': logs[-1].statusFailed,
				'paused': logs[-1].statusPaused,
			}

			# status chart
			charts['status'] = json.dumps({
				'type': 'doughnut',
				'data': {
					'datasets': [{
						'data': [
							logs[-1].statusOk,
							logs[-1].statusFailed,
							logs[-1].statusPaused,
						],
						'backgroundColor': [
							'#52ff52',
							'#ff5252',
							'#bdbdbd',
						],
					}],
					'labels': [
						'Passed',
						'Failed',
						'Paused'
					],
				},
				'options': {
					'legend': {
						'display': False,
					},
				},
			})

			# results chart
			charts['results'] = json.dumps({
			'type': 'bar',
				'data': {
					'datasets': [
						{
							'data': [x.statusOk for x in logs] + [None]*empty,
							'backgroundColor': '#52ff52',
							'label': 'Passed',
						},
						{
							'data': [x.statusFailed for x in logs] + [None]*empty,
							'backgroundColor': '#ff5252',
							'label': 'Failed',
						},
						{
							'data': [x.statusPaused for x in logs] + [None]*empty,
							'backgroundColor': '#bdbdbd',
							'label': 'Paused',
						},
					],
					'labels': [x.startTime.strftime('%Y-%m-%d %H:%M') for x in logs] + [None]*empty,
				},
				'options': {
					'legend': {
						'display': False,
					},
					'tooltips': {
						'mode': 'index',
						'intersect': False,
					},
					'scales': {
						'xAxes': [{
							'gridLines': {
								'display': False,
								'drawBorder': False,
							},
							'ticks': {
								'display': False,
							},
							'stacked': True,
						}],
						'yAxes': [{
							'gridLines': {
								'display': False,
								'drawBorder': False,
							},
							'ticks': {
								'display': False,
							},
							'stacked': True,
						}],
					},
				},
			})

			# duration chart
			charts['duration'] = json.dumps({
			'type': 'line',
				'data': {
					'datasets': [{
						'data': [str(x.duration) for x in logs] + [None]*empty,
						'fill': False,
						'borderColor': '#cfd8dc',
						'borderWidth': 1, 
						'pointBackgroundColor': 'transparent',
						'pointBorderColor': '#0277bd',
						'pointRadius': 5,
						'lineTension': 0,
						'label': 'Duration',
					}],
					'labels': [x.startTime.strftime('%Y-%m-%d %H:%M') for x in logs] + [None]*empty,
				},
				'options': {
					'legend': {
						'display': False,
					},
					'tooltips': {
						'mode': 'index',
						'intersect': False,
					},
					'layout': {
						'padding': {
							'top': 10,
							'right': 10,
						},
					},
					'scales': {
						'xAxes': [{
							'gridLines': {
								'display': False,
								'drawBorder': False,
							},
							'ticks': {
								'display': False,
							},
						}],
						'yAxes': [{
							'gridLines': {
								'display': False,
								'drawBorder': False,
							},
							'ticks': {
								'display': False,
							},
						}],
					},
				},
			})


			data['-'.join(name.split('.'))] = charts

		db.close()

		return data


	def show_dashboard(self):
		#
		# shows html dashboard
		#

		filename = os.path.join(os.path.dirname(__file__), self.path_to_reports, self.created.strftime('dashboard-%Y%m%d-%H%M%S.html'))
		self.generate_dashboard(filename)
		url = f'file://{filename}'
		webbrowser.open(url, new=2)

	def generate_dashboard(self, filename):
		file_loader = FileSystemLoader(os.path.join(os.path.dirname(__file__), self.path_to_reports, 'templates'))
		env = Environment(loader=file_loader)

		template = env.get_template('dashboard.html')

		with open(filename, 'w') as f:
			f.write(template.render(data=self.data, created=self.created.strftime('%Y-%m-%d %H:%M:%S')))


import logging
from xlrd import open_workbook
import itertools
import json
import baangt.base.CustGlobalConstants as CGC
import baangt.base.GlobalConstants as GC
from baangt.base.Utils import utils
import baangt.TestSteps.Exceptions
from pathlib import Path
import xl2dict
import re
from random import randint

logger = logging.getLogger("pyC")


class HandleDatabase:
    def __init__(self, linesToRead, globalSettings=None):
        self.lineNumber = 3
        # FIXME: This is still not clean. GlobalSettings shouldn't be predefined in CustomConstants-Class
        self.globals = {
            CGC.CUST_TOASTS: "",
            GC.EXECUTION_STAGE: "",
            GC.TESTCASEERRORLOG: "",
            CGC.VIGOGFNUMMER: "",
            CGC.SAPPOLNR: "",
            CGC.PRAEMIE: "",
            CGC.POLNRHOST: "",
            GC.TESTCASESTATUS: "",
            GC.TIMING_DURATION: "",
            GC.SCREENSHOTS: "",
            GC.TIMELOG: "",
        }
        if globalSettings:
            for setting, value in globalSettings.items():
                self.globals[setting] = value
        self.range = self.__buildRangeOfRecords(linesToRead)
        self.rangeDict = {}
        self.__buildRangeDict()
        self.df_json = None
        self.dataDict = []
        self.recordPointer = 0
        self.sheet_dict = {}

    def __buildRangeDict(self):
        """
        Interprets the Range and creates a DICT of values, that we can loop over later

        @return: Creates empty self.rangeDict
        """
        for lRangeLine in self.range:
            for x in range(lRangeLine[0], lRangeLine[1]+1):
                self.rangeDict[x] = ""

    def __buildRangeOfRecords(self, rangeFromConfigFile):
        lRange = []
        if not rangeFromConfigFile:
            # No selection - means all records
            return [[0,99999]]
        else:
            # Format: 4;6-99;17-200;203 or
            # Format: 4,6-100,800-1000
            if ";" in rangeFromConfigFile:
                for kombination in rangeFromConfigFile.split(";"):
                    lRange.append(HandleDatabase.__buildRangeOfRecordsOneEntry(kombination))
            elif "," in rangeFromConfigFile:
                for kombination in rangeFromConfigFile.split(","):
                    lRange.append(HandleDatabase.__buildRangeOfRecordsOneEntry(kombination))
            else:
                lRange.append(HandleDatabase.__buildRangeOfRecordsOneEntry(rangeFromConfigFile))

        # Make sure these are numbers:
        for lRangeLine in lRange:
            lRangeLine[0] = HandleDatabase.__sanitizeNumbers(lRangeLine[0])
            lRangeLine[1] = HandleDatabase.__sanitizeNumbers(lRangeLine[1])

        return lRange

    @staticmethod
    def __sanitizeNumbers(numberIn):
        if isinstance(numberIn, dict):
            numberIn = numberIn['default']
            try:
                  return int(numberIn.strip())
            except:
                  return 0
        numberIn = numberIn.strip()
        return int(numberIn)

    @staticmethod
    def __buildRangeOfRecordsOneEntry(rangeIn):
        if "-" in rangeIn:
            # This is a range (17-22)
            return HandleDatabase.__buildRangeOfRecordsSingleRange(rangeIn)
        else:
            # This is a single entry:
            return [rangeIn, rangeIn]

    @staticmethod
    def __buildRangeOfRecordsSingleRange(rangeIn):
        lSplit = rangeIn.split("-")
        return [lSplit[0], lSplit[1]]

    def read_excel(self, fileName, sheetName):
        fileName = utils.findFileAndPathFromPath(fileName)
        if not fileName:
            logger.critical(f"Can't open file: {fileName}")
            return

        book = open_workbook(fileName)
        sheet = book.sheet_by_name(sheetName)

        # read header values into the list
        keys = [sheet.cell(0, col_index).value for col_index in range(sheet.ncols)]

        for row_index in range(1, sheet.nrows):
            temp_dic = {}
            for col_index in range(sheet.ncols):
                temp_dic[keys[col_index]] = sheet.cell(row_index, col_index).value
                if type(temp_dic[keys[col_index]])==float:
                    temp_dic[keys[col_index]] = repr(temp_dic[keys[col_index]])
                    if temp_dic[keys[col_index]][-2:]==".0":
                        temp_dic[keys[col_index]] = temp_dic[keys[col_index]][:-2]

            self.dataDict.append(temp_dic)

        for temp_dic in self.dataDict:
            new_data_dic ={}
            for keys in temp_dic:
                if '$(' in str(temp_dic[keys]):
                    while '$(' in str(temp_dic[keys]):
                        start_index = temp_dic[keys].index('$(')
                        end_index = temp_dic[keys][start_index:].index(')')+start_index
                        data_to_replace_with = temp_dic[temp_dic[keys][start_index+2:end_index]]
                        temp_dic[keys] = temp_dic[keys].replace(
                            temp_dic[keys][start_index:end_index+1], data_to_replace_with
                        )
                    if temp_dic[keys][:4] == "RRD_":
                        rrd_string = self.__process_rrd_string(temp_dic[keys])
                        rrd_data = self.__rrd_string_to_python(rrd_string[4:], fileName)
                        for data in rrd_data:
                            new_data_dic[data] = rrd_data[data]
            for key in new_data_dic:
                temp_dic[key] = new_data_dic[key]

    def __compareEqualStageInGlobalsAndDataRecord(self, currentNewRecordDict:dict) -> bool:
        """
        As method name says, compares, whether Stage in Global-settings is equal to stage in Data Record,
        so that this record might be excluded, if it's for the wrong Stage.

        :param currentNewRecordDict: The current Record
        :return: Boolean
        """
        lAppend = True
        if self.globals.get(GC.EXECUTION_STAGE):
            if currentNewRecordDict.get(GC.EXECUTION_STAGE):
                if currentNewRecordDict[GC.EXECUTION_STAGE] != self.globals[GC.EXECUTION_STAGE]:
                    lAppend = False
        return lAppend

    def __processRrd(self, sheet_name, data_looking_for, data_to_match: dict):
        """
        For more detail please refer to TestDataGenerator.py
        :param sheet_name:
        :param data_looking_for:
        :param data_to_match:
        :return: dictionary of TargetData
        """
        matching_data = [list(x) for x in itertools.product(*[data_to_match[key] for key in data_to_match])]
        assert sheet_name in self.sheet_dict, f"Excel file doesn't contain {sheet_name} sheet. Please recheck. Called in 'RRD_'"
        base_sheet = self.sheet_dict[sheet_name]
        data_lis = []
        if type(data_looking_for) == str:
            data_looking_for = data_looking_for.split(",")

        for data in base_sheet:
            if len(matching_data) == 1 and len(matching_data[0]) == 0:
                if data_looking_for[0] == "*":
                    data_lis.append(data)
                else:
                    data_lis.append({keys: data[keys] for keys in data_looking_for})
            else:
                if [data[key] for key in data_to_match] in matching_data:
                    if data_looking_for[0] == "*":
                        data_lis.append(data)
                    else:
                        data_lis.append({keys: data[keys] for keys in data_looking_for})
        return data_lis

    def __rrd_string_to_python(self, raw_data, fileName):
        """
        Convert string to python data types
        :param raw_data:
        :return:
        """
        first_value = raw_data[1:-1].split(',')[0].strip()
        second_value = raw_data[1:-1].split(',')[1].strip()
        if second_value[0] == "[":
            second_value = ','.join(raw_data[1:-1].split(',')[1:]).strip()
            second_value = second_value[:second_value.index(']') + 1]
            third_value = [x.strip() for x in ']'.join(raw_data[1:-1].split(']')[1:]).split(',')[1:]]
        else:
            third_value = [x.strip() for x in raw_data[1:-1].split(',')[2:]]
        evaluated_list = ']],'.join(','.join(third_value)[1:-1].strip().split('],')).split('],')
        if evaluated_list[0] == "":
            evaluated_dict = {}
        else:
            evaluated_dict = {
                splited_data.split(':')[0]: self.__splitList(splited_data.split(':')[1]) for splited_data in
                evaluated_list
            }
        if second_value[0] == "[" and second_value[-1] == "]":
            second_value = self.__splitList(second_value)
        if first_value not in self.sheet_dict:
            self.sheet_dict, _ = self.__read_excel(path=fileName)
        processed_datas = self.__processRrd(first_value, second_value, evaluated_dict)
        assert len(processed_datas)>0, f"No matching data for RRD_. Please check the input file. Was searching for " \
                                       f"{first_value}, {second_value} and {str(evaluated_dict)} " \
                                       f"but didn't find anything"
        return processed_datas[randint(0, len(processed_datas)-1)]

    def __process_rrd_string(self, rrd_string):
        """
        For more detail please refer to TestDataGenerator.py
        :param rrd_string:
        :return:
        """
        processed_string = ','.join([word.strip() for word in rrd_string.split(', ')])
        match = re.match(
            r"(RRD_(\(|\[))[a-zA-z0-9\s]+,(\[?[a-zA-z\s,]+\]?|)|\*,\[([a-zA-z0-9\s]+:\[[a-zA-z0-9,\s]+\](,?))*\]",
            processed_string
        )
        err_string = f"{rrd_string} not matching pattern RRD_(sheetName,TargetData," \
                     f"[Header1:[Value1],Header2:[Value1,Value2]])"
        assert match, err_string
        return processed_string

    def __splitList(self, raw_data):
        """
        Will convert string list to python list.
        i.e. "[value1,value2,value3]" ==> ["value1","value2","value3"]
        :param raw_data: string of list
        :return: Python list
        """
        proccesed_datas = [data.strip() for data in raw_data[1:-1].split(",")]
        return proccesed_datas

    def __read_excel(self, path, sheet_name=""):
        """
        For more detail please refer to TestDataGenerator.py
        :param path: Path to raw data xlsx file.
        :param sheet_name: Name of base sheet sheet where main input data is located. Default will be the first sheet.
        :return: Dictionary of all sheets and data, Dictionary of base sheet.
        """
        wb = open_workbook(path)
        sheet_lis = wb.sheet_names()
        sheet_dict = {}
        for sheet in sheet_lis:
            xl_obj = xl2dict.XlToDict()
            data = xl_obj.fetch_data_by_column_by_sheet_name(path,sheet_name=sheet)
            sheet_dict[sheet] = data
        if sheet_name == "":
            base_sheet = sheet_dict[sheet_lis[0]]
        else:
            assert sheet_name in sheet_dict, f"Excel file doesn't contain {sheet_name} sheet. Please recheck."
            base_sheet = sheet_dict[sheet_name]
        return sheet_dict, base_sheet

    def readNextRecord(self):
        """
        We built self.range during init. Now we need to iterate over the range(s) in range,
        find appropriate record and return that - one at a time

        @return:
        """
        if len(self.rangeDict) == 0:
            # All records were processed
            return None
        try:
            # the topmost record of the RangeDict (RangeDict was built by the range(s) from the TestRun
            # - 1 because there's a header line in the Excel-Sheet.
            lRecord = self.dataDict[(list(self.rangeDict.keys())[0])]
            while not self.__compareEqualStageInGlobalsAndDataRecord(lRecord):
                logger.debug(f"Skipped record {str(lRecord)[:30]} due to wrong stage: {lRecord[GC.EXECUTION_STAGE]} vs. "
                             f"{self.globals[GC.EXECUTION_STAGE]}")
                self.rangeDict.pop(list(self.rangeDict.keys())[0])
                lRecord = self.dataDict[(list(self.rangeDict.keys())[0])]
        except Exception as e:
            logger.debug(f"Couldn't read record from database: {list(self.rangeDict.keys())[0]}")
            self.rangeDict.pop(list(self.rangeDict.keys())[0])
            return None

        # Remove the topmost entry from the rangeDict, so that next time we read the next entry in the lines above
        self.rangeDict.pop(list(self.rangeDict.keys())[0])
        return self.updateGlobals(lRecord)

    def readTestRecord(self, lineNumber=None):
        if lineNumber:
            self.lineNumber = lineNumber -1  # Base 0 vs. Base 1
        else:
            self.lineNumber += 1 # add 1 to read next line number

        try:
            record = self.dataDict[self.lineNumber]
            logger.info(f"Starting with Testrecord {self.lineNumber}, Details: " +
                        str({k: record[k] for k in list(record)[0:5]}))
            return self.updateGlobals(record)
        except Exception as e:
            logger.critical(f"Couldn't read record# {self.lineNumber}")

    def updateGlobals(self, record):

        self.globals[CGC.CUST_TOASTS] = ""
        self.globals[GC.TESTCASEERRORLOG] = ""
        self.globals[GC.TIMING_DURATION] = ""
        self.globals[GC.TIMELOG] = ""
        self.globals[GC.TESTCASESTATUS] = ""
        self.globals[GC.SCREENSHOTS] = ""

        record.update(self.globals)
        return record



import os
from selenium import webdriver
from appium import webdriver as Appiumwebdriver
from selenium.webdriver.support import expected_conditions as ec
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.firefox.options import Options as ffOptions
from selenium.common.exceptions import *
from selenium.webdriver.common import keys
from baangt.base import GlobalConstants as GC
from baangt.TestSteps import Exceptions
from baangt.base.Utils import utils
from baangt.base.BrowserHandling.BrowserHelperFunction import BrowserDriverData
from baangt.base.BrowserHandling.BrowserHelperFunction import BrowserHelperFunction as helper
import time
import logging
from pathlib import Path
import json
import sys

logger = logging.getLogger("pyC")

class WebdriverFunctions:
    """
    The webdriverclass for BrowserHandling to interact with selenium webdriver.
    """

    BROWSER_DRIVERS = {
        GC.BROWSER_FIREFOX: webdriver.Firefox,
        GC.BROWSER_CHROME: webdriver.Chrome,
        GC.BROWSER_SAFARI: webdriver.Safari,
        GC.BROWSER_EDGE: webdriver.Edge,
        GC.BROWSER_REMOTE: webdriver.Remote,
        GC.BROWSER_APPIUM : Appiumwebdriver.Remote
    }

    @staticmethod
    def getDownloadFolderFromProfile(profile: webdriver.FirefoxProfile):
        return profile.__getattribute__("default_preferences")["browser.download.dir"]

    @staticmethod
    def webdriver_setFirefoxProfile(browserProxy, randomProxy=None):
        profile = webdriver.FirefoxProfile()
        if browserProxy:
            profile.set_proxy(browserProxy.selenium_proxy())

        if randomProxy:
            """
            from selenium import webdriver
            return webdriver.Proxy({
                "httpProxy": self.proxy,
                "sslProxy": self.proxy,
            })
            """
            # We shall use a random Proxy from the list:
            PROXY = f"{randomProxy['ip']}:{randomProxy['port']}"

            logger.info(f"Using Proxy-Server: {PROXY}")

            ffProxy = webdriver.Proxy( {
                        "httpProxy":PROXY,
                        "ftpProxy":PROXY,
                        "sslProxy":PROXY,
                        "noProxy":None,
                        "proxy_type":"MANUAL",
                        "proxyType":"MANUAL",
                        "class":"org.openqa.selenium.Proxy",
                        "autodetect":False
                        })

            profile.set_proxy(ffProxy)

        profile.set_preference("browser.download.folderList", 2)
        profile.set_preference("browser.helperApps.alwaysAsk.force", False)
        profile.set_preference("browser.download.manager.showWhenStarting", False)
        profile.set_preference("browser.download.manager.showAlertOnComplete", False)
        profile.set_preference('browser.helperApps.neverAsk.saveToDisk',
                            'application/octet-stream,application/pdf,application/x-pdf,application/vnd.pdf,application/zip,application/octet-stream,application/x-zip-compressed,multipart/x-zip,application/x-rar-compressed, application/octet-stream,application/msword,application/vnd.ms-word.document.macroEnabled.12,application/vnd.openxmlformats-officedocument.wordprocessingml.document,application/vnd.ms-excel,application/pdf,application/vnd.openxmlformats-officedocument.spreadsheetml.sheet,application/vnd.openxmlformats-officedocument.spreadsheetml.sheet,application/vnd.openxmlformats-officedocument.wordprocessingml.document,application/vnd.openxmlformats-officedocument.spreadsheetml.sheet,application/rtf,application/vnd.openxmlformats-officedocument.spreadsheetml.sheet,application/vnd.ms-excel,application/vnd.ms-word.document.macroEnabled.12,application/vnd.openxmlformats-officedocument.wordprocessingml.document,application/xls,application/msword,text/csv,application/vnd.ms-excel.sheet.binary.macroEnabled.12,text/plain,text/csv/xls/xlsb,application/csv,application/download,application/vnd.openxmlformats-officedocument.presentationml.presentation,application/octet-stream')
        profile.set_preference('browser.helperApps.neverAsk.openFile',
                            'application/octet-stream,application/pdf,application/x-pdf,application/vnd.pdf,application/zip,application/octet-stream,application/x-zip-compressed,multipart/x-zip,application/x-rar-compressed, application/octet-stream,application/msword,application/vnd.ms-word.document.macroEnabled.12,application/vnd.openxmlformats-officedocument.wordprocessingml.document,application/vnd.ms-excel,application/pdf,application/vnd.openxmlformats-officedocument.spreadsheetml.sheet,application/vnd.openxmlformats-officedocument.spreadsheetml.sheet,application/vnd.openxmlformats-officedocument.wordprocessingml.document,application/vnd.openxmlformats-officedocument.spreadsheetml.sheet,application/rtf,application/vnd.openxmlformats-officedocument.spreadsheetml.sheet,application/vnd.ms-excel,application/vnd.ms-word.document.macroEnabled.12,application/vnd.openxmlformats-officedocument.wordprocessingml.document,application/xls,application/msword,text/csv,application/vnd.ms-excel.sheet.binary.macroEnabled.12,text/plain,text/csv/xls/xlsb,application/csv,application/download,application/vnd.openxmlformats-officedocument.presentationml.presentation,application/octet-stream')
        profile.set_preference("browser.download.dir", helper.browserHelper_setBrowserDownloadDirRandom())
        profile.set_preference("browser.download.manager.useWindow", False)
        profile.set_preference("browser.download.manager.focusWhenStarting", False)
        profile.set_preference("browser.download.manager.showAlertOnComplete", False)
        profile.set_preference("browser.download.manager.closeWhenDone", True)
        profile.set_preference("pdfjs.enabledCache.state", False)
        profile.set_preference("pdfjs.disabled", True) # This is nowhere on the Internet! But that did the trick!

        # Set volume to 0
        profile.set_preference("media.volume_scale", "0.0")
        return profile

    @staticmethod
    def webdriver_createBrowserOptions(browserName, desiredCapabilities, browserMobProxy=None, randomProxy=None):
        """
        Translates desired capabilities from the Testrun (or globals) into specific BrowserOptions for the
        currently active browser

        @param browserName: any of the GC.BROWSER*
        @param desiredCapabilities: Settings from TestRun or globals
        @param browserMobProxy: Proxy-Server IP+Port of internal BrowserMobProxy.
        @param randomProxy: Proxy-Server IP+Port of random external Proxy
        @return: the proper BrowserOptions for the currently active browser.
        """

        # Default Download Directory for Attachment downloads
        if browserName == GC.BROWSER_CHROME:
            lOptions = ChromeOptions()
            prefs = {"plugins.plugins_disabled" : ["Chrome PDF Viewer"],
                     "plugins.always_open_pdf_externally": True,
                     "profile.default_content_settings.popups": 0,
                     "download.default_directory": helper.browserHelper_setBrowserDownloadDirRandom(),  # IMPORTANT - ENDING SLASH V IMPORTANT
                     "directory_upgrade": True}
            lOptions.add_experimental_option("prefs", prefs)
            # Set Proxy for Chrome. First RandomProxy (External), if set. If not, then internal Browsermob
            if randomProxy:
                lOptions.add_argument(f"--proxy-server={randomProxy['ip']}:{randomProxy['port']}")
            elif browserMobProxy:
                lOptions.add_argument('--proxy-server={0}'.format(browserMobProxy.proxy))
        elif browserName == GC.BROWSER_FIREFOX:
            lOptions = ffOptions()
        else:
            lOptions = None

        if desiredCapabilities and lOptions:
            # sometimes instead of DICT comes a string with DICT-Format
            if isinstance(desiredCapabilities, str) and "{" in desiredCapabilities and "}" in desiredCapabilities:
                desiredCapabilities = json.loads(desiredCapabilities.replace("'", '"'))

            if isinstance(desiredCapabilities, dict):
                if desiredCapabilities.get(GC.BROWSER_MODE_HEADLESS):
                    logger.debug("Starting in Headless mode")
                    lOptions.headless = True
                else:
                    # statement not matched
                    pass
            else:
                # statement not matched
                pass
        else:
            # statement not matched
            pass

        return lOptions

    @staticmethod
    def getDownloadFolderFromChromeOptions(options: ChromeOptions):
        return options.__getattribute__("experimental_options")["prefs"]["download.default_directory"]

    @staticmethod
    def webdriver_getCurrentHTMLReference(driver):
        """
        Get a reference of the current HTML-Tag into self.html. We need that for stale check
        "waitForPageLoadAfterButtonClick"
        :return:
        """
        RETRY_COUNTER_LIMIT = 5
        e = None
        html = None
        retryCount = 0

        while retryCount < RETRY_COUNTER_LIMIT and not html:
            try:
                html = driver.find_element_by_tag_name('html')  # This is for waitForPageLoadAfterButton
            except NoSuchElementException as e:
                logger.debug(f"had a NoSuchElementException: {e}")
            except NoSuchWindowException as e:
                logger.debug(f"had a noSuchWindowException: {e}")
            except WebDriverException as e:
                logger.debug(f"had a WebDriverException: {e}")
            except BaseException as e:
                logger.warning(f"had an unknown exception (should be checked): {e}")

            retryCount += 1
            time.sleep(0.5)

        if retryCount >= RETRY_COUNTER_LIMIT:
            raise Exceptions.baangtTestStepException(f"Couldn't locate HTML element in Page. "
                                                     f"No idea what's going on. This was the last error"
                                                     f" (check logs for more): {e}")

        return html


    @staticmethod
    def webdriver_tryAndRetry(browserData, timeout=20, optional=False):
        """
        In: Locator
        Out: Boolean whether the element was found or not.

        Also sets the self.element for further use by other Methods (for instance to setText or read existing value)

        The method is resistant to common timing problems (can't work 100% of the time but will remove at least 80%
        of your pain compared to directly calling Selenium Methods).
        """
        REQUEST_TIMEOUT_MINIMUM = 1.5
        REQUEST_POLL_FREQUENCY = 0.5

        element = None
        html = None
        begin = time.time()
        elapsed = 0
        if timeout < REQUEST_TIMEOUT_MINIMUM:
            pollFrequency = timeout / 3
        else:
            pollFrequency = REQUEST_POLL_FREQUENCY

        internalTimeout = timeout / 5

        lLoopCount = 0

        try:
            html = WebdriverFunctions.webdriver_getCurrentHTMLReference(browserData.driver)
        except BaseException as e:
            raise Exceptions.baangtTestStepException(f"__getCurrentHTMLReference was not successful: {e}")

        while not element and elapsed < timeout:
            lLoopCount += 1
            try:
                driverWait = WebDriverWait(browserData.driver, timeout=internalTimeout, poll_frequency=pollFrequency)

                if By.ID == browserData.locatorType or By.CSS_SELECTOR == browserData.locatorType:
                    element = driverWait.until(ec.visibility_of_element_located((browserData.locatorType, browserData.locator)))
                elif By.CLASS_NAME == browserData.locatorType:
                    element = browserData.driver.find_element_by_class_name(browserData.locator)
                elif By.XPATH == browserData.locatorType:
                    # visibility of element sometimes not true, but still clickable. If we tried already
                    # 2 times with visibility, let's give it one more try with Presence of element
                    if lLoopCount > 1:
                        logger.debug(f"Tried 2 times to find visible element, now trying presence "
                                     f"of element instead, XPATH = {browserData.locator}")
                        element = driverWait.until(ec.presence_of_element_located((browserData.locatorType, browserData.locator)))
                    else:
                        element = driverWait.until(ec.visibility_of_element_located((browserData.locatorType, browserData.locator)))
            except StaleElementReferenceException as e:
                logger.debug("Stale Element Exception - retrying " + str(e))
                time.sleep(pollFrequency)
            except ElementClickInterceptedException as e:
                logger.debug("ElementClickIntercepted - retrying " + str(e))
                time.sleep(pollFrequency)
            except TimeoutException as e:
                logger.debug("TimoutException - retrying " + str(e))
                time.sleep(pollFrequency)
            except NoSuchElementException as e:
                logger.debug("Retrying Webdriver Exception: " + str(e))
                time.sleep(pollFrequency)
            except InvalidSessionIdException as e:
                logger.debug("WebDriver Exception - terminating testrun: " + str(e))
                raise Exceptions.baangtTestStepException
            except NoSuchWindowException as e:
                helper.browserHelper_log(logging.CRITICAL, "WebDriver Exception - terminating testrun: " + str(e), browserData)
                raise Exceptions.baangtTestStepException
            except ElementNotInteractableException as e:
                logger.debug("Most probably timeout exception - retrying: " + str(e))
                time.sleep(pollFrequency)
            except WebDriverException as e:
                helper.browserHelper_log(logging.ERROR, "Retrying WebDriver Exception: " + str(e), browserData)
                time.sleep(2)

            elapsed = time.time() - begin

        return element, html


    @staticmethod
    def webdriver_doSomething(command, element, value=None, timeout=20, optional=False, browserData=None):
        """
        Will interact in an element (that was found before by findBy-Method and stored in self.element) as defined by
        ``command``.

        Command can be "SETTEXT" (GC.CMD_SETTEXT), "CLICK" (GC.CMD_CLICK), "FORCETEXT" (GC.CMD_FORCETEXT).

        Similarly to __try_and_retry the method is pretty robust when it comes to error handling of timing issues.

        """
        NUMBER_OF_SEND_KEY_BACKSPACE = 10
        COUNTER_LIMIT_RETRY = 2
        COUNTER_LIMIT_ELEMENT_REF = 4
        COUNTER_LIMIT_ELEMENT_NOT_INTERACT = 5

        didWork = False
        elapsed = 0
        counter = 0
        begin = time.time()

        while not didWork and elapsed < timeout:
            counter += 1
            logger.debug(f"__doSomething {command} with {value}")
            try:
                if command.upper() == GC.CMD_SETTEXT:
                    if not value:
                        value = ""
                    element.send_keys(value)
                elif command.upper() == GC.CMD_CLICK:
                    element.click()
                elif command.upper() == GC.CMD_FORCETEXT:
                    element.clear()
                    for i in range(0, NUMBER_OF_SEND_KEY_BACKSPACE):
                        element.send_keys(keys.Keys.BACKSPACE)
                    time.sleep(0.1)
                    element.send_keys(value)
                didWork = True
            except ElementClickInterceptedException as e:
                logger.debug("doSomething: Element intercepted - retry")
                time.sleep(0.2)
            except StaleElementReferenceException as e:
                logger.debug(f"doSomething: Element stale - retry {browserData.locatorType} {browserData.locator}")
                # If the element is stale after 2 times, try to re-locate the element
                if counter < COUNTER_LIMIT_RETRY:
                    time.sleep(0.2)
                elif counter < COUNTER_LIMIT_ELEMENT_REF:
                    element, begin = WebdriverFunctions.webdriver_refindElementAfterError(browserData, timeout)
                else:
                    raise Exceptions.baangtTestStepException(e)
            except NoSuchElementException as e:
                logger.debug("doSomething: Element not there yet - retry")
                time.sleep(0.5)
            except InvalidSessionIdException as e:
                helper.browserHelper_log(logging.ERROR, f"Invalid Session ID Exception caught - aborting... {e} ", browserData)
                raise Exceptions.baangtTestStepException(e)
            except ElementNotInteractableException as e:
                if counter < COUNTER_LIMIT_RETRY:
                    logger.debug(f"Element not interactable {browserData.locatorType} {browserData.locator}, retrying")
                    time.sleep(0.2)
                elif counter < COUNTER_LIMIT_ELEMENT_NOT_INTERACT:
                    logger.debug(f"Element not interactable {browserData.locatorType} {browserData.locator}, re-finding element")
                    element, begin = WebdriverFunctions.webdriver_refindElementAfterError(browserData, timeout)
                else:
                    helper.browserHelper_log(logging.ERROR, f"Element not interactable {e}", browserData)
                    raise Exceptions.baangtTestStepException(e)
            except NoSuchWindowException as e:
                raise Exceptions.baangtTestStepException(e)
            elapsed = time.time() - begin

        if not didWork:
            if optional:
                logger.debug(
                    f"Action not possible after {timeout} s, Locator: {browserData.locatorType}: {browserData.locator}, but flag 'optional' is set")
            else:
                raise Exceptions.baangtTestStepException(f"Action not possible after {timeout} s, Locator: {browserData.locatorType}: {browserData.locator}")
        else:
            # Function successful
            pass

        return didWork

    @staticmethod
    def webdriver_refindElementAfterError(browserData, timeout):
        element, _ = WebdriverFunctions.webdriver_tryAndRetry(browserData, timeout=timeout / 2, optional=True)
        if element:
            logger.debug(f"Re-Found element {browserData.locatorType}: {browserData.locator}, will retry ")
            begin = time.time()
        else:
            raise Exceptions.baangtTestStepException(
                f"Element {browserData.locatorType} {browserData.locator} couldn't be found. "
                f"Tried to re-find it, but not element was not found")
        return begin, element


import setuptools

if __name__ == '__main__':
    with open("README.md", "r") as fh:
        long_description = fh.read()

    setuptools.setup(
        name="baangt",
        version="1.1.1",
        author="Bernhard Buhl",
        author_email="info@baangt.org",
        description="Open source Test Automation Suite for MacOS, Windows, Linux",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://baangt.org",
        packages=setuptools.find_packages(),
        data_files=[('baangt', ["baangt/ressources/baangtLogo.png", "baangt/ressources/baangtLogo2020Small.png"])],
        package_data={"baangt.ui.pyqt": ['globalSetting.json',]},
        install_requires=["Appium-Python-Client", "beautifulsoup4", "browsermob-proxy",
                          "dataclasses", "dataclasses-json",
                          "faker",  "gevent", "lxml",
                          "openpyxl",
                          "Pillow", "pluggy", "pyperclip",  "pyQT5",
                          "requests", "requests-toolbelt",
                          "schwifty", "selenium", "sqlalchemy",
                          "urllib3",
                          "xl2dict", "xlrd", "xlsxwriter",
                           ],
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        include_package_data=True,
        python_requires='>=3.6',
    )



import baangt.base.GlobalConstants as GC
import requests
from requests.exceptions import ConnectionError
from logging import getLogger
from base64 import b64encode
from dataclasses import dataclass, asdict, field

logger = getLogger("pyC")


@dataclass
class PDFCompareDetails:
    fileName: str = field(default="")
    referenceID: str = field(default="")
    BLOB: str = field(default="")
    Status: str = field(default="")
    StatusText: str = field(default="")
    BLOB_OUT: str = field(default="")
    ResultText: str = field(default="")
    Description: str = field(default="No description available from calling functionality")
    newUUID: str = field(default="")


class PDFCompare:
    def __init__(self, baseURL="http://localhost:5000/"):
        self.callParam = None
        self.baseURL = baseURL

    def compare_multiple(self, files: dict):
        """

        :param files: a Dict:
                KEY = PDF-Filename and path
                "referenceID" of the original file to compare to
                "BLOB": The file as base64-encoded. Provide that in case it's not reachable from this class. If BLOB is
                       given, the file will not be read.
                "Status": OK, NOK
                "StatusText": Explanation in case of NOK
                "BLOB_OUT": If Status was NOK, this is the base64-encoded result PDF with marked differences.
                "ResultText": If Status = NOK, the text differences between the two PDFs
        :return:
        """
        results = {}

        for file, details in files.items():
            if isinstance(details, PDFCompareDetails):
                lDetails = details
            else:
                try:
                    details["newUUID"] = ""                                   # NewUUID shouldn't be set
                    lDetails = PDFCompareDetails(str(file), **details)        # Unpack DICT into the Dataclass
                except TypeError as e:
                    details["Status"] = "NOK"
                    details["StatusText"] = f"Wrong call to function. Technical error. One or more elements missing: {e}"
                    files[file] = details
                    continue

            logger.debug(f"Comparing PDF-Files: {lDetails.fileName} with reference {lDetails.referenceID}")
            lDetails = self.__callService(file, lDetails)

            if isinstance(details, PDFCompareDetails):
                files[file] = lDetails
            else:
                files[file] = asdict(lDetails)

        return files

    def uploadNewReferenceFile(self, details: PDFCompareDetails):
        """
        Will upload the file specified in details.Filename and receive the new UUID
        :param details: filled in dataclass PDFCompareDetails
        :return: same structure. If everything went right, with a newUUID of the uploaded file.
        """

        details = self.__callService(file=None, details=details, typeOfService="createOriginal")
        return details

    def __callService(self, file : str, details: PDFCompareDetails, typeOfService="compare"):
        if not file:
            lFile = details.fileName
        else:
            lFile = file

        blobToCompare = self.__getBlobFromInput(lFile, details.BLOB)

        if not blobToCompare:
            details.Status = "NOK"
            details.StatusText = f"No input BLOB given and file {lFile} not there or can't be read."
            return details
        else:
            details.BLOB = blobToCompare

        if typeOfService == 'compare':
            details = self.__callComparisonService(details=details)
        elif typeOfService == 'createOriginal':
            details = self.__callUploadService(details=details, endpoint="/upload_reference")
        else:
            details.Status = "NOK"
            details.StatusText = f"Unknown typeOfService specified: {typeOfService}"

        return details

    def __callComparisonService(self, details: PDFCompareDetails):
        """
            Step 1: Call Upload of new BLOB
            Step 2: Call Comparison service after having received the new BLOB
        :param details:
        :return:
        """

        details = self.__callUploadService(details)

        if details.newUUID:
            details = self.__callComparisonServiceExecution(details)

        return details

    def __callUploadService(self, details: PDFCompareDetails, endpoint="/upload_original"):
        if endpoint == '/upload_original':
            files = {
                "original": (details.fileName, details.BLOB),
                "reference_uuid": (None, details.referenceID),
                "description": (None, details.Description)
            }
        elif endpoint == '/upload_reference':
            files = {
                "reference": (details.fileName, details.BLOB),
                "description": (None, details.Description)
            }

        lResponse = self.__executeRequest(params=None, methodGetOrPost="post", endpoint=endpoint, files=files)

        if not isinstance(lResponse, requests.Response):
            details.Status = "NOK"
            details.StatusText = f"Error when calling remote service. Config wrong? Server down? Error was: {lResponse}"
            return details

        if lResponse.status_code == 200:
            details.Status = "OK"
            lJson = lResponse.json()
            details.newUUID = lJson[0].get("uuid")
            if not details.newUUID:
                details.Status = "NOK"
                details.StatusText = f"Response didn't have UUID. Here's the response: {lResponse.text}"
        else:
            details.Status = "NOK"
            details.StatusText = f"Error {lResponse.status_code} from Request to Service. " \
                                 f"Error from service was {lResponse.text}"

        return details

    def __callComparisonServiceExecution(self, details: PDFCompareDetails):
        # Requests Comparison of the two IDs, typically after they were uploaded
        # UUID 1 = new ID of this file
        # UUID 2 = Reference ID of old file (Uploaded as a reference and given as parameter in the Testcase!
        params = {"uuid1": details.newUUID,
                  "uuid2": details.referenceID}
        lResponse = self.__executeRequest(endpoint='/comparison', params=params, methodGetOrPost="get")

        if lResponse.status_code == 200:
            details.Status = "OK"
            lJson = lResponse.json()
            if lJson["result"] == "OK":
                details.Status = "OK"
                details.StatusText = ""
            else:
                details.Status = "NOK"
                details.StatusText = "Diff in Original: " + lJson["result"]["orig_file_diff"] + \
                                     "\nDiff in Reference:" +  lJson["result"]["orig_file_diff"]
        else:
            details.Status = "NOK"
            details.StatusText = f"Error {lResponse.status_code} from Request to Service. " \
                                 f"Error from service was {lResponse.text}"

        return details

    def __executeRequest(self, endpoint, params, methodGetOrPost="get", files=None):
        lUrl = f"{self.baseURL}/{endpoint}"
        if methodGetOrPost.lower() == "post":
            try:
                lResponse = requests.post(url=lUrl, params=params, files=files)
            except ConnectionError as e:
                return e
            except Exception as e:
                logger.critical(f"New uncought exception. Should be looked into! Exception was: {e}")
                return e

        elif methodGetOrPost.lower() == "get":
            try:
                lResponse = requests.get(url=lUrl, params=params)
            except ConnectionError as e:
                return e
            except Exception as e:
                logger.critical(f"New uncought exception. Should be looked into! Exception was: {e}")
                return e
        else:
            logger.critical(f"called with wrong method: {methodGetOrPost}")

        return lResponse

    def __getBlobFromInput(self, file, blob):
        if blob:
            return blob

        lReadBinary = None

        try:
            blob = open(file, "rb").read()
            # blob = b64encode(blob)
        except FileNotFoundError as e:
            logger.critical(f"File {file} not found to read current PDF for comparison")
            return None
        except Exception as e:
            logger.critical(f"File {file} could not be decoded properly. Weird stuff though.")

        return blob
